@@ -2,7 +2,6 @@
 #define SNA_DAMAGE_H
 
 #include <regionstr.h>
-#include <list.h>
 
 #include "compiler.h"
 
@@ -216,6 +215,19 @@ static inline int sna_damage_contains_box(struct sna_damage *damage,
 
 	return _sna_damage_contains_box(damage, box);
 }
+static inline int sna_damage_contains_box__offset(struct sna_damage *damage,
+						  const BoxRec *box, int dx, int dy)
+{
+	BoxRec b;
+
+	if (DAMAGE_IS_ALL(damage))
+		return PIXMAN_REGION_IN;
+
+	b = *box;
+	b.x1 += dx; b.x2 += dx;
+	b.y1 += dy; b.y2 += dy;
+	return _sna_damage_contains_box(damage, &b);
+}
 bool _sna_damage_contains_box__no_reduce(const struct sna_damage *damage,
 					const BoxRec *box);
 static inline bool
@@ -259,20 +271,22 @@ static inline void sna_damage_reduce_all(struct sna_damage **_damage,
 	if (damage == NULL || DAMAGE_IS_ALL(damage))
 		return;
 
-	if (damage->mode == DAMAGE_ADD &&
-	    damage->extents.x1 <= 0 &&
-	    damage->extents.y1 <= 0 &&
-	    damage->extents.x2 >= width &&
-	    damage->extents.y2 >= height) {
-		if (damage->dirty) {
-			damage = *_damage = _sna_damage_reduce(damage);
-			if (damage == NULL)
-				return;
-		}
+	if (damage->mode == DAMAGE_ADD) {
+		if (damage->extents.x1 <= 0 &&
+		    damage->extents.y1 <= 0 &&
+		    damage->extents.x2 >= width &&
+		    damage->extents.y2 >= height) {
+			if (damage->dirty) {
+				damage = *_damage = _sna_damage_reduce(damage);
+				if (damage == NULL)
+					return;
+			}
 
-		if (damage->region.data == NULL)
-			*_damage = _sna_damage_all(damage, width, height);
-	}
+			if (damage->region.data == NULL)
+				*_damage = _sna_damage_all(damage, width, height);
+		}
+	} else
+		*_damage = _sna_damage_reduce(damage);
 }
 
 void __sna_damage_destroy(struct sna_damage *damage);
